/*
ARM7 code for RAC

Copyright 2007 Damian Yerrick <pinoandchester@pineight.com>

This work is provided 'as-is', without any express or implied
warranty.  In no event will the authors be held liable for any
damages arising from the use of this work.

Permission is granted to anyone to use this work for any purpose,
including commercial applications, and to alter it and redistribute
it freely, subject to the following restrictions:

1. The origin of this work must not be misrepresented; you must
   not claim that you wrote the original work. If you use
   this work in a product, an acknowledgment in the product
   documentation would be appreciated but is not required.
2. Altered source versions must be plainly marked as such, and must
   not be misrepresented as being the original work.
3. This notice may not be removed or altered from any source
   distribution.

"Source" is the preferred form of a work for making changes to it.

*/
/*
Based on libnds ARM7 template by wintermute
*/

#include <nds.h>
#include <stdlib.h>
#include "talkback.h"

//////////////////////////////////////////////////////////////////////


#define TOUCH_CAL_X1 (*(vs16*)0x027FFCD8)
#define TOUCH_CAL_Y1 (*(vs16*)0x027FFCDA)
#define TOUCH_CAL_X2 (*(vs16*)0x027FFCDE)
#define TOUCH_CAL_Y2 (*(vs16*)0x027FFCE0)
#define SCREEN_WIDTH    256
#define SCREEN_HEIGHT   192
s32 TOUCH_WIDTH;
s32 TOUCH_HEIGHT;
s32 TOUCH_OFFSET_X;
s32 TOUCH_OFFSET_Y;


//////////////////////////////////////////////////////////////////////


extern const unsigned short midi2psgFreq[120];

#define C(octave)  (octave * 12 +  0)
#define CS(octave) (octave * 12 +  1)
#define D(octave)  (octave * 12 +  2)
#define DS(octave) (octave * 12 +  3)
#define E(octave)  (octave * 12 +  4)
#define F(octave)  (octave * 12 +  5)
#define FS(octave) (octave * 12 +  6)
#define G(octave)  (octave * 12 +  7)
#define GS(octave) (octave * 12 +  8)
#define A(octave)  (octave * 12 +  9)
#define AS(octave) (octave * 12 + 10)
#define B(octave)  (octave * 12 + 11)



const unsigned char pentatonic[5] = {1, 4, 6, 8, 11};

unsigned int heartbeat = 0;
int axePhase = 0;
int axeNextFreq = -1;
int axeLastPan = 0;
int axeLastDuty = 0;
int axeNextDuty = 0;

void axe(int touchXpx, int touchYpx) {
  int tempoCounter = heartbeat % 17 % 9;

  /*
   * Horizontal (X) = pitch and panning
   * Vertical (Y) = duty cycle
   */
  if(touchXpx > 255)
    touchXpx = 255;
  if(touchXpx > 0) {
    axeNextFreq = touchXpx >> 3;
    axeNextDuty = (touchYpx / 48) << 24;
  } else {
    axeNextFreq = -1;
  }

  // If it's time to start a note, do so
  if(tempoCounter == 0) {
    if(axeNextFreq >= 0) {
      axePhase = 7;
      axeLastDuty = axeNextDuty;
      axeLastPan = SOUND_PAN(4 * axeNextFreq + 2);
      int octave = axeNextFreq / 5 + 2;
      int semitone = pentatonic[axeNextFreq % 5] + 3;
      SCHANNEL_TIMER(12) = midi2psgFreq[octave * 12 + semitone];
    }
  }

  // Update the note
  if(axePhase >= 0) {
    int vol = (axePhase * axePhase * 3) >> 1;
    SCHANNEL_CR(12) = SCHANNEL_ENABLE | SOUND_FORMAT_PSG | axeLastPan
                   | axeLastDuty | vol;
    --axePhase;
  }
}


static unsigned int chNote[3], chTime[3];


/*
The following is "Take Me Out" popularized by Franz Ferdinand.
They copied Mr. Resetti's song from ACPG.
Commercial users might want to change the following data tables.
*/

static const unsigned int startTime[3] = {32, 28, 18};

static const signed char bassMain[64] = {
  E(3), 0, -1, 0, E(3), 0, -1, 0, E(3), 0, -1, 0, E(3), 43, 0, 0, 
  E(3), 0, -1, 0, E(3), 0, -1, 0, E(3), 0, -1, 0, E(3), G(3), 0, 0, 
  A(2), 0, -1, 0, A(2), 0, -1, 0, A(2), 0, -1, 0, A(2), AS(2), 0, 0, 
  B(2), 0, -1, 0, B(2), 0, -1, 0, B(2), 0, -1, 0, B(2), D(3), 0, 0
};

static const signed char ch2Main[64] = {
  -1, 0, G(5), 0, E(5), 0, E(5), D(5), B(4), 0, D(5), 0, E(5), 0, 0, 0,
  -1, 0, G(5), 0, E(5), 0, E(5), D(5), B(4), 0, D(5), 0, E(5), 0, 0, 0,
  -1, 0, G(5), 0, E(5), 0, E(5), D(5), B(4), 0, D(5), 0, E(5), 0, 0, 0,
  -1, 0, FS(5), FS(5), FS(5), FS(5), FS(5), FS(5),
  FS(5), FS(5), FS(5), FS(5), D(5), 0, D(5), 0
};

static const signed char ch3Main[64] = {
  -1, 0, D(5), D(5), D(5), D(5), D(5), -1,
  D(5), D(5), D(5), E(5), -1, 0, E(5), E(5),
  -1, 0, D(5), D(5), D(5), D(5), D(5), -1,
  D(5), D(5), D(5), E(5), -1, 0, E(5), E(5),
  -1, 0, C(5), C(5), C(5), C(5), C(5), -1,
  C(5), C(5), C(5), C(5), -1, 0, C(5), C(5),
  -1, 0, D(5), D(5), D(5), D(5), D(5), -1,
  D(5), D(5), D(5), D(5), A(4), G(4), E(4), E(4)
};

static const signed char ch3Alt[64] = {
  -1, 0, FS(5), G(5), B(5), 0, 0, 0, FS(5), G(5), B(5), 0, E(5), 0, 0, 0,
  -1, 0, FS(5), G(5), B(5), 0, 0, 0, FS(5), G(5), B(5), 0, E(5), 0, 0, 0,
  -1, 0, FS(5), G(5), B(5), 0, 0, 0, FS(5), G(5), B(5), 0, E(5), 0, 0, 0,
  -1, 0, FS(5), G(5), D(6), 0, 0, 0, FS(5), G(5), D(6), 0, E(5), 0, 0, 0
};

static const signed char *const scripts[4][3] = {
  { bassMain, ch2Main, ch3Main },
  { bassMain, ch2Main, ch3Main },
  { bassMain, NULL, ch3Alt },
  { bassMain, NULL, ch3Alt }
};

/*
End "Take Me Out"
*/

int order = -1;

/*
 * Plays a repeating drum track and bass track.
 */
void runSound(void) {
  unsigned int tempoCounter = heartbeat % (68 * 8);
  unsigned int row = tempoCounter * 2 / 17;
  unsigned int tick = (tempoCounter * 2 % 17) / 2;

  if (row == 0 && tick == 0) {
    ++order;
    if (order == 4) {
      order = 0;
    }
  }
  for (int ch = 0; ch < 3; ++ch) {
    const signed char *script = scripts[order][ch];
    int note = script ? script[row] : 0;
    if (note < 0) {
      chTime[ch] = 0;
    } else if (note > 0 && tick == 0) {
      chTime[ch] = startTime[ch];
      chNote[ch] = midi2psgFreq[note];
    } else if (chTime[ch] > 0) {
      --chTime[ch];
    }
  }
  
  int snareFill = tempoCounter > 68 * 8 - 17;
  int crashCymbal = tempoCounter < 34;

  tempoCounter %= 68;
  tempoCounter = (tempoCounter * 56 + 34) / 68;

  if ((tempoCounter >= 28 && tempoCounter < 42)
      || snareFill) {
    int snareCounter = snareFill
                       ? tempoCounter % 7
                       : tempoCounter - 28;
    // play snare
    int snareVol = ((17 - snareCounter) * (17 - snareCounter)) >> 3;

    SCHANNEL_CR(14) = SCHANNEL_ENABLE | SOUND_FORMAT_PSG | SOUND_PAN(72)
                   | snareVol;
    SCHANNEL_TIMER(14) = SOUND_FREQ(10923);
  } else if (crashCymbal) {
    SCHANNEL_CR(14) = SCHANNEL_ENABLE | SOUND_FORMAT_PSG | SOUND_PAN(64)
                   | (28 - tempoCounter);
    SCHANNEL_TIMER(14) = SOUND_FREQ(32768);
  } else {
    // play hat
    int hatPhase = tempoCounter;
    if(hatPhase >= 28)
      hatPhase -= 28;
    if(hatPhase >= 14)
      hatPhase -= 14;
    else
      hatPhase *= 8;
    if(hatPhase >= 14)
      hatPhase = 0;
    else
      hatPhase = 14 - hatPhase;
    hatPhase <<= 1;

    SCHANNEL_CR(14) = SCHANNEL_ENABLE | SOUND_FORMAT_PSG | SOUND_PAN(64)
                   | hatPhase;
    SCHANNEL_TIMER(14) = SOUND_FREQ(32768);
  }

  // play kick and bass
  {
    int kickPhase = tempoCounter;
    if(kickPhase >= 28)
      kickPhase -= 28;
    if(kickPhase >= 14) {
      SCHANNEL_CR(11) = SCHANNEL_ENABLE | SOUND_FORMAT_PSG | SOUND_PAN(64)
                     | 0x03000000;
    } else {

      // Play kick drum
      int kickVol = (kickPhase >= 8) ? 0 : 10 * (8 - kickPhase);
      SCHANNEL_CR(11) = SCHANNEL_ENABLE | SOUND_FORMAT_PSG | SOUND_PAN(64)
                     | 0x03000000 | kickVol;
      SCHANNEL_TIMER(11) = SOUND_FREQ(660 * 8) * (kickPhase + 4);
    }
  }
  
  // Play bass
  SCHANNEL_CR(8) = SCHANNEL_ENABLE | SOUND_FORMAT_PSG | SOUND_PAN(64)
                   | 0x01000000 | (chTime[0] * 2);
  SCHANNEL_TIMER(8) = chNote[0];
  SCHANNEL_CR(9) = SCHANNEL_ENABLE | SOUND_FORMAT_PSG | SOUND_PAN(64)
                   | 0x00000000 | (chTime[1] * 2);
  SCHANNEL_TIMER(9) = chNote[1];
  SCHANNEL_CR(10) = SCHANNEL_ENABLE | SOUND_FORMAT_PSG | SOUND_PAN(64)
                   | 0x01000000 | (chTime[2] * 2);
  SCHANNEL_TIMER(10) = chNote[2];
  
}


/**
 * Sets the power management chip on the ARM7 SPI bus.
 */
__attribute__((noinline)) void PM_SetRegister(int reg, int control)
{

  SerialWaitBusy();

  REG_SPICNT = SPI_ENABLE | SPI_DEVICE_POWER | SPI_BAUD_1MHz | SPI_CONTINUOUS;
  REG_SPIDATA = reg;

  SerialWaitBusy();

  REG_SPICNT = SPI_ENABLE | SPI_DEVICE_POWER | SPI_BAUD_1MHz;
  REG_SPIDATA = control;
}

void powerOffDS(void)
{
  PM_SetRegister(0, 1 << 6);//6 DS power (0: on, 1: shut down!)
}



int vcount;
touchPosition first,tempPos;

//---------------------------------------------------------------------------------
void VcountHandler() {
//---------------------------------------------------------------------------------
	static int lastbut = -1;
	
	uint16 but=0, x=0, y=0, xpx=0, ypx=0, z1=0, z2=0;

	but = REG_KEYXY;

	if (!( (but ^ lastbut) & (1<<6))) {
 
		tempPos = touchReadXY();

		if ( tempPos.x == 0 || tempPos.y == 0 ) {
			but |= (1 <<6);
			lastbut = but;
		} else {
			x = tempPos.x;
			y = tempPos.y;
			xpx = tempPos.px;
			ypx = tempPos.py;
			z1 = tempPos.z1;
			z2 = tempPos.z2;
		}
		
	} else {
		lastbut = but;
		but |= (1 <<6);
	}

	if ( vcount == 80 ) {
		first = tempPos;
	} else {
		if (	abs( xpx - first.px) > 10 || abs( ypx - first.py) > 10 ||
				(but & ( 1<<6)) ) {

			but |= (1 <<6);
			lastbut = but;

		} else { 	
			IPC->mailBusy = 1;
			IPC->touchX			= x;
			IPC->touchY			= y;
			IPC->touchXpx		= xpx;
			IPC->touchYpx		= ypx;
			IPC->touchZ1		= z1;
			IPC->touchZ2		= z2;
			IPC->mailBusy = 0;
		}
	}
	IPC->buttons		= but;
	vcount ^= (80 ^ 130);
	SetYtrigger(vcount);

}

//---------------------------------------------------------------------------------
void VblankHandler(void) {
//---------------------------------------------------------------------------------

    volatile P8A7Talkback *tb = (P8A7Talkback *)IPC->soundData;
    int axeFreq = -1;
   
    if (tb) {
      if (tb->needPowerOff) {
        tb->needPowerOff = 0;
        powerOFF(POWER_SOUND);
        powerOffDS();
      }
      if (tb->needNote) {
        axeFreq = tb->needNote * 8;
      }
    }

    axe(axeFreq, 48);
    runSound();

    ++heartbeat;
 
}

//---------------------------------------------------------------------------------
int main(int argc, char ** argv) {
//---------------------------------------------------------------------------------

	IPC->soundData = 0;
	rtcReset();
    rtcGetTimeAndDate((uint8 *)&(IPC->time.rtc.year));

	//enable sound
	powerON(POWER_SOUND);
	SOUND_CR = SOUND_ENABLE | SOUND_VOL(0x7F);

	irqInit();
	irqSet(IRQ_VBLANK, VblankHandler);
	SetYtrigger(80);
	vcount = 80;
	irqSet(IRQ_VCOUNT, VcountHandler);
	irqEnable(IRQ_VBLANK | IRQ_VCOUNT);

	// Keep the ARM7 idle
	while (1) swiWaitForVBlank();
}

//////////////////////////////////////////////////////////////////////
